# -*- coding: utf-8 -*-
"""Test the raw anonymization.

For each supported file format, implement a test.
"""
# Authors: Alex Rockhill <aprockhill@mailbox.org>
#
# License: BSD (3-clause)

import os.path as op
import mne
from mne.datasets import testing

from numpy.testing import assert_array_almost_equal
from mne.utils import _TempDir

import ephys_anonymizer

basepath = op.join(op.dirname(ephys_anonymizer.__file__), 'tests', 'data')

data_path = testing.data_path()
fif_fname = op.join(data_path, 'MEG', 'sample',
                    'sample_audvis_trunc_raw.fif')
edf_fname = op.join(data_path, 'EDF', 'test_reduced.edf')

raw_fif = mne.io.read_raw_fif(fif_fname)
raw_edf = mne.io.read_raw_edf(edf_fname)


def test_raw_anonymize():
    out_dir = _TempDir()
    out_fname = op.join(out_dir, 'test-anon-raw.fif')
    _ = ephys_anonymizer.raw_anonymize(fif_fname, out_fname=out_fname)
    raw = mne.io.read_raw_fif(out_fname)
    assert raw.info['meas_date'].year == 2000
    assert raw.info['meas_date'].month == 1
    assert raw.info['meas_date'].day == 1
    assert raw.info['experimenter'] == 'mne_anonymize'
    assert_array_almost_equal(raw.get_data(), raw.get_data(), decimal=10)

    _ = ephys_anonymizer.raw_anonymize(edf_fname, out_fname, overwrite=True)
    raw = mne.io.read_raw_fif(out_fname)
    assert_array_almost_equal(raw.get_data(), raw.get_data(), decimal=10)



"""Command Line Interface for video anonymization."""
# Authors: Alex Rockhill <aprockhill@mailbox.org>
#
# License: BSD (3-clause)
import argparse

import ephys_anonymizer


def video_anonymize():
    """Run video_anonymize command."""
    parser = argparse.ArgumentParser()
    parser.add_argument('filename', type=str,
                        help='Name of the video file to anonymize. '
                        'If there are any errors be sure to enter '
                        'the full path as the first troubleshooting option')
    parser.add_argument('out_fname', nargs='?', default=None, type=str,
                        help='Filename to save out to')
    parser.add_argument('--scale', default=10, type=int, required=False,
                        help='How many neighboring pixels to use, '
                             'try scaling up or down if faces are not '
                             'being found')
    parser.add_argument('--show', default=False, type=bool, required=False,
                        help='Whether to show the anonymized video')
    parser.add_argument('--verbose', default=True, type=bool,
                        required=False,
                        help='Set verbose output to True or False.')
    parser.add_argument('--overwrite', default=False, type=bool,
                        required=False, help='Whether to overwrite')
    args = parser.parse_args()
    ephys_anonymizer.video_anonymize(args.filename, out_fname=args.out_fname,
                                     scale=args.scale, show=args.show,
                                     verbose=args.verbose,
                                     overwrite=args.overwrite)


def raw_anonymize():
    """Run raw_anonymize command."""
    parser = argparse.ArgumentParser()
    parser.add_argument('filename', type=str,
                        help='Name of the raw file to anonymize')
    parser.add_argument('out_fname', nargs='?', default=None, type=str,
                        help='Filename to save out to')
    parser.add_argument('--verbose', default=True, type=bool,
                        required=False,
                        help='Set verbose output to True or False.')
    parser.add_argument('--overwrite', default=False, type=bool,
                        required=False, help='Whether to overwrite')
    args = parser.parse_args()
    ephys_anonymizer.raw_anonymize(args.filename, out_fname=args.out_fname,
                                   verbose=args.verbose,
                                   overwrite=args.overwrite)



# -*- coding: utf-8 -*-
"""Test the video anonymization.

For each supported file format, implement a test.
"""
# Authors: Alex Rockhill <aprockhill@mailbox.org>
#
# License: BSD (3-clause)

import os.path as op
import cv2
from mne.utils import _TempDir

import ephys_anonymizer

basepath = op.join(op.dirname(ephys_anonymizer.__file__), 'tests', 'data')

face_data = dict(x=list(), y=list())
with open(op.join(basepath, 'face_data.tsv'), 'r') as fid:
    _ = fid.readline()
    for line in fid:
        x, y = line.rstrip().split('\t')
        face_data['x'].append(int(float(x)))
        face_data['y'].append(int(float(y)))


def test_video_anonymize():
    out_dir = _TempDir()
    for fname in ('test_vid.mp4', 'test_vid.mov', 'test_vid.avi'):
        ext = op.splitext(fname)[-1]
        out_fname = ephys_anonymizer.video_anonymize(
            op.join(out_dir, fname), show=False, overwrite=True)
        cap = cv2.VideoCapture(out_fname)
        i = 0
        ret, frame = cap.read()
        while ret:
            ret, frame = cap.read()
            # > 7 because mov, mp4 are imprecise
            if any(frame[face_data['y'][i], face_data['x'][i]] > 7):
                raise ValueError('Face not anonymized for {}, frame {}, '
                                 'value {}'.format(ext, i,
                                                   frame[face_data['y'][i],
                                                         face_data['x'][i]]))
            ret, frame = cap.read()
            i += 1
        cap.release()
    cv2.destroyAllWindows()



# -*- coding: utf-8 -*-
"""Anonymize a video.

Anonymize a video with a black box over any faces.
"""
# Authors: Alex Rockhill <aprockhill@mailbox.org>
#
# License: BSD (3-clause)

import sys
import os.path as op


# based on https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials
# /py_objdetect/py_face_detection/py_face_detection.html
def _find_face_and_cover(frame_gray, frame_color, cascades, scale, verbose):
    """Find faces and cover with black."""
    # get the locations of the faces
    faces = cascades[0].detectMultiScale(frame_gray, 1.1, scale)
    if len(faces) == 0:
        if verbose:
            print('No faces found trying second cascades')
        faces = cascades[1].detectMultiScale(frame_gray, 1.1, scale)
        if len(faces) == 0:
            faces = cascades[2].detectMultiScale(frame_gray, 1.1, scale)
    elif len(faces) > 1 and verbose:
        sys.stdout.write('{} faces found in frame'.format(len(faces)))
    for (x, y, w, h) in faces:
        # select the areas where the face was found
        frame_color[y: y + h, x:x + w] = 0
    return frame_color


def video_anonymize(fname, out_fname=None, scale=10, show=False,
                    verbose=True, overwrite=False):
    """Anonymize a video.

    This function will use the Viola-Jones algorithm to detect faces
    in a video and put a black box where the face is. The video is
    saved as the file name with anon added in the same directory.

    Parameters
    ----------
    fname : str
        The full file path of the video file. Currently,
        '.mov', '.mp4' and '.avi' are tested.
    out_fname : str
        The file name to save the anonymized video out to.
        Defaults to fname with '-anon.avi' after.
    scale : int
        Number of close neighbors to require. Increase if too many
        false positive faces in videos.
    show : bool
        Whether to show the new anonymized video.
        Defaults to False.
    verbose : bool
        Set verbose output to True or False.
    overwrite : bool
        Whether to overwrite the existing file.
        Defaults to False.

    Returns
    -------
    out_fname : str
        The name of the anonymized video file.
    """
    import cv2
    basename, ext = op.splitext(fname)
    if out_fname is None:
        out_fname = '{}-anon.avi'.format(basename)
    else:
        out_basename, out_ext = op.splitext(out_fname)
        out_fname = out_basename + '.avi'
    if op.isfile(out_fname) and not overwrite:
        raise ValueError('Anonymized file exists, use '
                         '`overwrite=True` to overwrite')
    if verbose:
        print('Reading in {}'.format(fname))
    cascades = [cv2.CascadeClassifier('{}{}.xml'.format(
                cv2.data.haarcascades, name)) for name in
                ('haarcascade_frontalface_default',
                 'haarcascade_profileface',
                 'haarcascade_frontalface_alt_tree',
                 'haarcascade_frontalface_alt',
                 'haarcascade_frontalface_alt2')]
    cap = cv2.VideoCapture(fname)
    fps = cap.get(cv2.CAP_PROP_FPS)
    if ext == '.mov':
        frame_width = int(cap.get(4))
        frame_height = int(cap.get(3))
    else:
        frame_width = int(cap.get(3))
        frame_height = int(cap.get(4))

    out = cv2.VideoWriter(out_fname, cv2.VideoWriter_fourcc(*'MJPG'),
                          fps, (frame_width, frame_height))
    ret, frame = cap.read()
    if verbose:
        sys.stdout.write('Anonymizing .')
        sys.stdout.flush()
    while ret:
        if ext == '.mov':
            frame = frame.swapaxes(0, 1)
            frame = frame[:, ::-1]
        frame_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        frame = _find_face_and_cover(frame_gray, frame, cascades, scale,
                                     verbose)
        out.write(frame)
        if show:
            cv2.imshow('frame', frame)
            k = cv2.waitKey(30) & 0xff
            if k == 27:
                break
        ret, frame = cap.read()
        if verbose:
            sys.stdout.write('.')
            sys.stdout.flush()
    cap.release()
    out.release()
    cv2.destroyAllWindows()
    if verbose:
        print('\nVideo saved to {}'.format(out_fname))
    return out_fname


def raw_anonymize(fname, out_fname=None, verbose=True, overwrite=False):
    """Anonymize a raw file.

    This function uses the mne-python anonymize functions to
    anonymize raw data or a MATLAB script from Burke Rosen 2019-05-14
    at USCD to anonymize edf files only (not recommended).

    Parameters
    ----------
    fname : str
        The full file path of the raw file. Currently,
        '.fif'
    out_fname : str
        The file name to save the anonymized raw file out to.
        Defaults to fname with '-anon-raw.fif' after.
    verbose : bool
        Set verbose output to True or False.
    overwrite : bool
        Whether to overwrite the existing file.
        Defaults to False.

    Returns
    -------
    out_fname : str
        The name of the anonymized video file.
    """
    import mne
    basename, ext = op.splitext(fname)
    if out_fname is None:
        out_fname = '{}-anon-raw.fif'.format(basename)
    else:
        out_basename, out_ext = op.splitext(out_fname)
        if out_basename[-4:] in ('-raw', '_raw'):
            out_fname = out_basename + '.fif'
        else:
            out_fname = out_basename + '-raw.fif'
    if op.isfile(out_fname) and not overwrite:
        raise ValueError('Anonymized file exists, use '
                         '`overwrite=True` to overwrite')
    if verbose:
        print('Reading in {}'.format(fname))
    if ext == '.fif':
        raw = mne.io.read_raw_fif(fname, preload=False)
    elif ext == '.edf':
        raw = mne.io.read_raw_edf(fname, preload=False)
    elif ext == 'bdf':
        raw = mne.io.read_raw_bdf(fname, preload=False)
    elif ext == '.vhdr':
        raw = mne.io.read_raw_brainvision(fname, preload=False)
    elif ext == '.set':
        raw = mne.io.read_raw_eeglab(fname, preload=False)
    else:
        raise ValueError('Extension {} not recognized, options are'
                         'fif, edf, bdf, vhdr (brainvision) and set '
                         '(eeglab)'.format(ext))
    if verbose:
        print('Anonymizing')
    raw.anonymize()
    if verbose:
        print('Saving to {}'.format(out_fname))
    raw.save(out_fname, overwrite=overwrite)
    return out_fname



#! /usr/bin/env python
"""Setup ephys_anonymizer."""
import os
from setuptools import setup, find_packages

# get the version
version = None
with open(os.path.join('ephys_anonymizer', '__init__.py'), 'r') as fid:
    for line in (line.strip() for line in fid):
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().strip('\'')
            break
if version is None:
    raise RuntimeError('Could not determine version')


descr = ('ephys_anonymizer: Use the Viola-Jones algorithm to anonymize '
         'faces with a black box and anonymize electrophysiology '
         'data with mne-python')

DISTNAME = 'ephys_anonymizer'
DESCRIPTION = descr
MAINTAINER = 'Alex Rockhill'
MAINTAINER_EMAIL = 'aprockhill@mailbox.org'
URL = 'https://github.com/alexrockhill/ephys_anonymizer/'
LICENSE = 'BSD (3-clause)'
DOWNLOAD_URL = 'https://github.com/alexrockhill/ephys_anonymizer.git'
VERSION = version

if __name__ == "__main__":
    setup(name=DISTNAME,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=VERSION,
          download_url=DOWNLOAD_URL,
          long_description=open('README.rst').read(),
          long_description_content_type='text/x-rst',
          python_requires='~=3.5',
          classifiers=[
              'Intended Audience :: Science/Research',
              'Intended Audience :: Developers',
              'License :: OSI Approved',
              'Programming Language :: Python',
              'Topic :: Software Development',
              'Topic :: Scientific/Engineering',
              'Operating System :: Microsoft :: Windows',
              'Operating System :: POSIX',
              'Operating System :: Unix',
              'Operating System :: MacOS',
              'Programming Language :: Python :: 3.5',
              'Programming Language :: Python :: 3.6',
              'Programming Language :: Python :: 3.7',
          ],
          platforms='any',
          packages=find_packages(),
          entry_points={'console_scripts': [
              'video_anonymize = '
              'ephys_anonymizer.commands.run:video_anonymize',
              'raw_anonymize = '
              'ephys_anonymizer.commands.run:raw_anonymize'
          ]},
          project_urls={
              'Bug Reports':
                  'https://github.com/alexrockhill/ephys_anonymizer/issues',
              'Source': 'https://github.com/alexrockhill/ephys_anonymizer',
          },
          )
